<?php
/**
 * @file
 * Provides a Sphinx-based service class for the Search API.
 */

/**
 * Search service class using Sphinx RT indexes.
 */
class SearchApiSphinxService extends SearchApiAbstractService {
  // Dummy empty value for Multi-valued attributes.
  // Necessary as we can not filter on real empty value.
  protected $emptyMVA = 0x7FFFFFFF;

  /**
   * Overrides __construct().
   */
  public function __construct(SearchApiServer $server) {
    parent::__construct($server);
  }

  /**
   * Overrides configurationForm().
   */
  public function configurationForm(array $form, array &$form_state) {
    // Sphinx paths. Application user (ex:apache) must have read/write access.
    $form['path_structure'] = array(
      '#type' => 'fieldset',
      '#title' => t('Sphinx directory settings'),
      '#tree' => FALSE,
    );

    // Server base directory. App user (ex:apache) must have full access.
    $default = '';
    if (function_exists('posix_getuid')) {
      $process_user = posix_getpwuid(posix_geteuid());
      $app_user = $process_user['name'];
    }
    else {
      $app_user = get_current_user();
    }
    $form['path_structure']['base_directory'] = array(
      '#type' => 'textfield',
      '#title' => t('Base directory'),
      '#description' => t("Base directory in which the server and indexes files and directories will be created. The directory must exist and the user running the webserver(%user) must have full access. There is no default as this directory has to be created or its permissions set up by the user.", array(
        '%user' => $app_user,
      )),
      '#required' => TRUE,
      '#default_value' => $this->getOption('base_directory', $default),
      '#parents' => array('options', 'form', 'base_directory'),
    );

    // Sphinx paths. Application user (ex:apache) must have read/write access.
    $form['daemon_settings'] = array(
      '#type' => 'fieldset',
      '#title' => t('Sphinx daemon settings'),
      '#tree' => FALSE,
    );

    // Sphinx daemon host.
    $default = SEARCH_API_SPHINX_SERVER_DEFAULT_HOST;
    $form['daemon_settings']['host'] = array(
      '#type' => 'textfield',
      '#title' => t('Host'),
      '#description' => t('Host to which searchd will listen for this server. Default is %default. See !see.', array(
        '%default' => $default,
        '!see' => l(t('Sphinx documentation - Host'), 'http://sphinxsearch.com/docs/current.html#conf-listen'),
      )),
      '#required' => TRUE,
      '#default_value' => $this->getOption('host', $default),
      '#parents' => array('options', 'form', 'host'),
    );

    // Sphinx daemon port.
    $default = SEARCH_API_SPHINX_SERVER_DEFAULT_PORT;
    $form['daemon_settings']['port'] = array(
      '#type' => 'textfield',
      '#title' => t('Port'),
      '#description' => t('Port to which searchd will listen for this server. Default is %default. See !see.', array(
        '%default' => $default,
        '!see' => l(t('Sphinx documentation - Port'), 'http://sphinxsearch.com/docs/current.html#conf-listen'),
      )),
      '#required' => TRUE,
      '#default_value' => $this->getOption('port', $default),
      '#parents' => array('options', 'form', 'port'),
    );

    // Sphinx daemon path.
    $default = SEARCH_API_SPHINX_SERVER_DEFAULT_SEARCHD;
    $form['daemon_settings']['searchd'] = array(
      '#type' => 'textfield',
      '#title' => t('Sphinx daemon path'),
      '#description' => t("Path of the searchd daemon. Default is %default. Check your sphinx installation.", array(
        '%default' => $default,
      )),
      '#required' => TRUE,
      '#default_value' => $this->getOption('searchd', $default),
      '#parents' => array('options', 'form', 'searchd'),
    );

    // Sphinx server timezone.
    $default = date_default_timezone_get();
    $form['daemon_settings']['timezone'] = array(
      '#type' => 'textfield',
      '#title' => t('Timezone'),
      '#description' => t("Timezone used to start the deamon (UNIX only). Default is PHP default timezone: %default. See php.ini.", array(
        '%default' => $default,
      )),
      '#required' => TRUE,
      '#default_value' => $this->getOption('timezone', $default),
      '#parents' => array('options', 'form', 'timezone'),
    );

    // Sphinx daemon host.
    $default = SEARCH_API_SPHINX_SERVER_DEFAULT_RT_FLUSH_PERIOD;
    $form['daemon_settings']['rt_flush_period'] = array(
      '#type' => 'textfield',
      '#title' => t('Flush period'),
      '#description' => t('Periodically flushes the indexes from RAM to disk. Default is %default. See !see.', array(
        '%default' => $default,
        '!see' => l(t('Sphinx documentation - RT flush period'), 'http://sphinxsearch.com/docs/current.html#conf-rt-flush-period'),
      )),
      '#required' => TRUE,
      '#default_value' => $this->getOption('rt_flush_period', $default),
      '#parents' => array('options', 'form', 'rt_flush_period'),
    );

    // Search settings.
    $form['search_settings'] = array(
      '#type' => 'fieldset',
      '#title' => t('Sphinx search settings'),
      '#tree' => FALSE,
    );

    // Maximum of return matches.
    $default = SEARCH_API_SPHINX_SERVER_DEFAULT_MAX_MATCHES;
    $form['search_settings']['max_matches'] = array(
      '#type' => 'textfield',
      '#title' => t('Max matches'),
      '#description' => t('Maximum number of matches that can be returned. Default is %default. See !see.', array(
        '%default' => $default,
        '!see' => l(t('Sphinx documentation - Max matches'), 'http://sphinxsearch.com/docs/current.html#conf-max-matches'),
      )),
      '#required' => TRUE,
      '#default_value' => $this->getOption('max_matches', $default),
      '#parents' => array('options', 'form', 'max_matches'),
    );

    // Collation server.
    $default = SEARCH_API_SPHINX_SERVER_DEFAULT_COLLATION_SERVER;
    $form['search_settings']['collation_server'] = array(
      '#type' => 'textfield',
      '#title' => t('Collation server'),
      '#description' => t("Specifies the default collation used for incoming requests. Default is %default. See !see.", array(
        '%default' => $default,
        '!see' => l(t('Sphinx documentation - Collation server'), 'http://sphinxsearch.com/docs/current.html#conf-collation-server'),
      )),
      '#required' => TRUE,
      '#default_value' => $this->getOption('collation_server', $default),
      '#parents' => array('options', 'form', 'collation_server'),
    );

    // Keywords expansion limit.
    $default = SEARCH_API_SPHINX_SERVER_DEFAULT_EXPANSION_LIMIT;
    $form['search_settings']['expansion_limit'] = array(
      '#type' => 'textfield',
      '#title' => t('Expansion limit'),
      '#description' => t('The maximum number of expanded keywords for a single wildcard when wildcard (star *) is enabled. Default is %default. See !see.', array(
        '%default' => $default,
        '!see' => l(t('Sphinx documentation - Expansion limit'), 'http://sphinxsearch.com/docs/current.html#conf-expansion-limit'),
      )),
      '#required' => TRUE,
      '#default_value' => $this->getOption('expansion_limit', $default),
      '#parents' => array('options', 'form', 'expansion_limit'),
    );

    // Sphinx maximum number of concurrent searches.
    $default = SEARCH_API_SPHINX_SERVER_DEFAULT_MAX_CHILDREN;
    $form['daemon_settings']['max_children'] = array(
      '#type' => 'textfield',
      '#title' => t('Max children'),
      '#description' => t('Maximum number of concurrent searches. Default is %default. See !see.', array(
        '%default' => $default,
        '!see' => l(t('Sphinx documentation - Max children'), 'http://sphinxsearch.com/docs/current.html#conf-max-children'),
      )),
      '#required' => TRUE,
      '#default_value' => $this->getOption('max_children', $default),
      '#parents' => array('options', 'form', 'max_children'),
    );

    // TODO: additional searchd configuration options.
    return $form;
  }

  /**
   * Overrides configurationFormValidate().
   */
  public function configurationFormValidate(array $form, array &$values, array &$form_state) {
    // Base directory.
    if (!empty($values['base_directory']) && !is_writable($values['base_directory'])) {
      form_set_error('options][form][base_directory', t('You must enter a valid path with write permissions as Base directory.'));
    }

    // Daemon IP address.
    if (filter_var($values['host'], FILTER_VALIDATE_IP) === FALSE) {
      form_set_error('options][form][host', t('You must enter a valid IP address for the Sphinx daemon.'));
    }

    // Daemon Port.
    if (filter_var($values['port'], FILTER_VALIDATE_INT, array('options' => array('min_range' => 0, 'max_range' => 65535))) === FALSE) {
      form_set_error('options][form][port', t('You must enter a valid Port (between 0 and 65535) for the Sphinx daemon.'));
    }

    // Searchd path.
    if (!file_exists($values['searchd'])) {
      form_set_error('options][form][searchd', t('Daemon file (searchd) could not be found. Enter a valid path.'));
    }

    // Timezone.
    if (!in_array($values['timezone'], timezone_identifiers_list())) {
      form_set_error('options][form][timezone', t('You must enter a timezone as returned by timezone_identifiers_list().'));
    }

    // RT flush period.
    if (filter_var($values['rt_flush_period'], FILTER_VALIDATE_INT, array('options' => array('min_range' => 0, 'max_range' => 0x7FFFFFFF))) === FALSE) {
      form_set_error('options][form][rt_flush_period', t('You must enter a value superior or equal to 0 as Flush period.'));
    }

    // Max matches.
    if (filter_var($values['max_matches'], FILTER_VALIDATE_INT, array('options' => array('min_range' => 1, 'max_range' => 0x7FFFFFFF))) === FALSE) {
      form_set_error('options][form][max_matches', t('You must enter a positive integer as Max matches.'));
    }

    // Collation server.
    $servers = array('libc_ci', 'libc_cs', 'utf8_general_ci', 'binary');
    if (!in_array($values['collation_server'], $servers)) {
      form_set_error('options][form][collation_server', t('You must enter either libc_ci, libc_cs, utf8_general_ci or binary as collation server.'));
    }

    // Expansion limit.
    if (filter_var($values['expansion_limit'], FILTER_VALIDATE_INT, array('options' => array('min_range' => 0, 'max_range' => 1000))) === FALSE) {
      form_set_error('options][form][expansion_limit', t('You must enter a positive integer between 1 and 1000 for the Expansion limit option.'));
    }

    // Max children.
    if (filter_var($values['max_children'], FILTER_VALIDATE_INT, array('options' => array('min_range' => 0, 'max_range' => 1000))) === FALSE) {
      form_set_error('options][form][max_children', t('You must enter a positive integer between 1 and 1000 for the Max children option.'));
    }
  }

  /**
   * Overrides supportsFeature().
   */
  public function supportsFeature($feature) {
    $supported = drupal_map_assoc(array(
      'search_api_facets',
      'search_api_facets_operator_or',
      'search_api_mlt',
    ));
    return isset($supported[$feature]);
  }

  /**
   * Overrides postCreate().
   */
  public function postCreate() {
    // Create the server directory.
    $this->createSphinxServerDirectory();
  }

  /**
   * Overrides postUpdate().
   */
  public function postUpdate() {
    if ($this->server->options != $this->server->original->options) {
      // Update the sphinx server and flag for re-indexing if necessary.
      return $this->updateSphinxServer();
    }
    return FALSE;
  }

  /**
   * Overrides preDelete().
   */
  public function preDelete() {
    // Stop the daemon.
    $this->startStopSphinxDaemon('stop');

    // Delete the server directory.
    $this->deleteSphinxServerDirectory();
  }

  /**
   * Overrides viewSettings().
   */
  public function viewSettings() {
    $output = array();

    // Add option to start/stop daemon.
    if (!empty($this->server->enabled)) {
      $form_state = array();
      $form_state['server'] = $this;
      $output['status'] = drupal_build_form('search_api_sphinx_server_daemon_form', $form_state);
    }

    // Display settings.
    $form = $form_state = array();
    $option_form = $this->configurationForm($form, $form_state);
    $option_form['#title'] = t('Sphinx server settings');

    $element = $this->parseOptionFormElement($option_form, 'options');
    if (!empty($element)) {
      $settings = '';
      foreach ($element['option'] as $sub_element) {
        $settings .= $this->viewSettingElement($sub_element);
      }

      $output['settings'] = array(
        '#type' => 'fieldset',
        '#title' => $element['label'],
      );

      $output['settings'][] = array(
        '#type' => 'markup',
        '#markup' => '<div class="sphinx-server-settings">' . $settings . '</div>',
      );
    }

    return $output;
  }

  /**
   * Helper function. Parse an option form element.
   */
  protected function parseOptionFormElement($element, $key) {
    $children_keys = element_children($element);

    if (!empty($children_keys)) {
      $children = array();
      foreach ($children_keys as $child_key) {
        $child = $this->parseOptionFormElement($element[$child_key], $child_key);
        if (!empty($child)) {
          $children[] = $child;
        }
      }
      if (!empty($children)) {
        return array(
          'label' => isset($element['#title']) ? $element['#title'] : $key,
          'option' => $children,
        );
      }
    }
    elseif (isset($this->options[$key])) {
      return array(
        'label' => isset($element['#title']) ? $element['#title'] : $key,
        'option' => $key,
      );
    }

    return array();
  }

  /**
   * Helper function. Display a setting element.
   */
  protected function viewSettingElement($element) {
    $output = '';

    if (is_array($element['option'])) {
      $value = '';
      foreach ($element['option'] as $sub_element) {
        $value .= $this->viewSettingElement($sub_element);
      }
    }
    else {
      $value = $this->getOption($element['option']);
      $value = nl2br(check_plain(print_r($value, TRUE)));
    }
    $output .= '<dt><em>' . check_plain($element['label']) . '</em></dt>' . "\n";
    $output .= '<dd>' . $value . '</dd>' . "\n";

    return "<dl>\n{$output}</dl>";
  }

  /**
   * Overrides addIndex().
   */
  public function addIndex(SearchApiIndex $index) {
    // Update the server.
    $this->updateSphinxServer();
  }

  /**
   * Overrides fieldsUpdated().
   */
  public function fieldsUpdated(SearchApiIndex $index) {
    // Update the server. Remove index files for the given index.
    $this->updateSphinxServer($index);

    // Flag for re-indexing.
    return TRUE;
  }

  /**
   * Overrides removeIndex().
   */
  public function removeIndex($index) {
    // Update the server. Remove index files for the given index.
    $this->updateSphinxServer($index);
  }

  /**
   * Overrides indexItems().
   */
  public function indexItems(SearchApiIndex $index, array $items) {
    if (empty($items)) {
      return array();
    }

    $return_id_list = array();

    $index_name = $this->getSphinxIndexName($index);

    $max_matches = $this->getOption('max_matches', 1000);

    $select_all_keyword = "'" . $this->getSphinxSelectAllKeyword($index) . "'";

    // Fields (with mandatory ID field (unique) and dummy select all field).
    $field_id_list = array('id', 'sphinx_all');

    // Index fields.
    $index_fields = $this->getSphinxIndexFields($index);
    foreach ($index_fields as $field_id => $field_info) {
      $field_id_list[] = $field_info['sphinx_field_name'];
    }
    $field_number = count($field_id_list);

    $field_list = implode(',', $field_id_list);

    // Split the process into chunks ($max_matches items per chunk).
    foreach (array_chunk($items, $max_matches, TRUE) as $items_chunk) {
      // Get and sanitize the list of item ids.
      $ids = SearchApiSphinxService::sanitizeIDs(array_keys($items_chunk));
      // Get existing items from sphinx index.
      if (!empty($ids)) {
        $limit = count($ids);
        $ids = implode(',', $ids);
        // SphinxQL query. See querySphinx notes.
        $query = sprintf("SELECT id FROM %s WHERE id IN (%s) LIMIT %u", $index_name, $ids, $limit);
        $existing = $this->querySphinx($query)->fetchCol();
      }
      else {
        $existing = array();
      }

      // Get values to insert.
      $field_value_list = array('update' => array(), 'insert' => array());

      foreach ($items_chunk as $item_id => $item_fields) {
        // Sanitize item ID.
        $id = SearchApiSphinxService::sanitizeID($item_id);
        if (is_null($id)) {
          break;
        }

        // Item id and dummy select all keyword.
        $values = array($id, $select_all_keyword);

        foreach ($index_fields as $field_id => $field_info) {
          // Prepare and sanitize value to prevent SQL injection.
          $value = $this->convertFieldValueToSphinxValue($field_info['sphinx_field_type'], $item_fields[$field_id]['value']);

          // If invalid (NULL) we don't index the item.
          if (is_null($value)) {
            break;
          }

          // Array = multi-valued attribute.
          if (is_array($value)) {
            $value = '(' . implode(',', $value) . ')';
          }

          $values[] = $value;
        }

        // Only index if the number of values matches the number of fields.
        if (count($values) === $field_number) {
          $field_value_list[in_array($item_id, $existing) ? 'update' : 'insert'][] = '(' . implode(',', $values) . ')';
          $return_id_list[] = $item_id;
        }
      }

      // Insert or update given items.
      if (!empty($field_value_list['insert'])) {
        $value_list = implode(',', $field_value_list['insert']);
        // SphinxQL query. See querySphinx notes.
        $query = sprintf("INSERT INTO %s (%s) VALUES %s", $index_name, $field_list, $value_list);
        $result = $this->querySphinx($query)->rowCount();
      }
      if (!empty($field_value_list['update'])) {
        $value_list = implode(',', $field_value_list['update']);
        // SphinxQL query. See querySphinx notes.
        $query = sprintf("REPLACE INTO %s (%s) VALUES %s", $index_name, $field_list, $value_list);
        $result = $this->querySphinx($query)->rowCount();
      }
    }

    return $return_id_list;
  }

  /**
   * Overrides deleteItems().
   */
  public function deleteItems($ids = 'all', SearchApiIndex $index = NULL) {
    if (!isset($index)) {
      // Update server. Remove index files for all indexes.
      $this->updateSphinxServer($this->getIndexes());
    }
    elseif ($ids === 'all') {
      // Update server. Remove index files for the given index.
      $this->updateSphinxServer($index);
    }
    elseif (!empty($ids)) {
      $index_name = $this->getSphinxIndexName($index);

      // Sanitize IDs.
      $ids = SearchApiSphinxService::sanitizeIDs($ids);

      if (!empty($ids)) {
        $ids = implode(',', $ids);
        // SphinxQL query. See querySphinx notes.
        $query = sprintf("DELETE FROM %s WHERE id IN (%s)", $index_name, $ids);
        $result = $this->querySphinx($query)->rowCount();
      }
    }
  }

  /**
   * Overrides search().
   */
  public function search(SearchApiQueryInterface $query) {
    // Get the index.
    $index = $query->getIndex();

    // Get the sphinx index name.
    $index_name = $this->getSphinxIndexName($index);

    // Get the fields.
    $index_fields = $this->getSphinxIndexFields($index);

    // Map sphinx field names to their original ids.
    $sphinx_fields = array();
    foreach ($index_fields as $field_id => $field_info) {
      $sphinx_fields[$field_info['sphinx_field_name']] = $field_id;
    }

    // Search result.
    $search_result = array();

    // Maximum of returned matches.
    $max_matches = $this->getOption('max_matches', 1000);

    // Query options.
    $query_options = $query->getOptions();

    // Range.
    $query_offset = empty($query_options['offset']) ? 0 : $query_options['offset'];
    $query_limit = empty($query_options['limit']) ? $max_matches : $query_options['limit'];
    if ($query_limit > $max_matches || $query_limit < 0) {
      $query_limit = $max_matches;
    }

    // Limit.
    $search_query_limit = array($query_offset, $query_limit);

    // Search query options.
    $search_query_options = array();

    // Search query option 'field_weights'. Handles boosts for text fields.
    $field_weights = array();
    foreach ($index_fields as $field) {
      if (isset($field['boost']) && $field['sphinx_field_type'] === 'rt_field') {
        $field_weights[] = sprintf('%s=%d', $field['sphinx_field_name'], $field['boost'] * 10);
      }
    }
    if (!empty($field_weights)) {
      $search_query_options['field_weights'] = '(' . implode(',', $field_weights) . ')';
    }

    // Search query option 'ranker'.
    $search_query_options['ranker'] = 'proximity_bm25';
    $search_query_options['max_matches'] = $max_matches;

    // Full text queries.
    $search_query_full_text = array();

    // Select fields.
    $search_query_fields = array('*, weight() AS score');

    // Expressions.
    $search_query_expressions = array();

    // Filters.
    $search_query_filters = array();

    // Sort criteria.
    $search_query_sort = $this->getSphinxQueryOrderBy($query->getSort(), $index_fields);

    // Group by.
    $search_query_group_by = array();


    // SEARCH API MORE LIKE THIS.
    $mlt = $query->getOption('search_api_mlt');
    if (!empty($mlt['id']) && !empty($mlt['fields'])) {
      $entity_id = SearchApiSphinxService::sanitizeID($mlt['id']);
      if (is_null($entity_id)) {
        continue;
      }

      $entity_type = $index->item_type;

      // First add a filter so that this entity is not caught.
      $search_query_filters[] = sprintf("id <> %u", $entity_id);

      // Load the entity.
      $entities = entity_load($entity_type, array($entity_id));

      if (isset($entities[$entity_id])) {
        // Get the corresponding entity wrapper so we can get the field values.
        $entity_wrapper = entity_metadata_wrapper($entity_type, $entities[$entity_id]);

        // Extract the fields used to get related entities.
        if (isset($entity_wrapper)) {
          $mlt_fields = array();
          foreach ($mlt['fields'] as $field_id) {
            if (isset($index_fields[$field_id])) {
              $mlt_fields[$field_id] = $index_fields[$field_id];
            }
          }

          // Get the full fields with their values.
          $mlt_fields = search_api_extract_fields($entity_wrapper, $mlt_fields);

          // Get index full text fields.
          $index_full_text_fields = $index->getFulltextFields();

          // Get the MLT full text fields and their value.
          $mlt_full_text_fields = array();

          foreach ($mlt_fields as $field_id => $field_info) {
            if (in_array($field_id, $index_full_text_fields)) {
              unset($mlt_fields[$field_id]);
              $mlt_full_text_fields[$field_id] = $field_info['value'];
            }
          }

          // Create the appropriate MLT full text queries.
          if (!empty($mlt_full_text_fields)) {
            // More Like This options.
            $mlt_word_proximity = $this->getIndexOption($index, 'mlt', 'word_proximity', 0);
            $mlt_min_word_len = $this->getIndexOption($index, 'mlt', 'min_word_len', 1);

            // Keyword separator: either 'OR' or 'NEAR' (proximity).
            $keyword_separator = $mlt_word_proximity == 0 ? '|' : " NEAR/{$mlt_word_proximity} ";

            // Function used to filter keywords by length.
            $filter_keywords = function ($keyword) use ($mlt_min_word_len) {
              return drupal_strlen($keyword) >= $mlt_min_word_len;
            };

            // Go through all the text fields to search for.
            foreach ($mlt_full_text_fields as $field_id => $field_value) {
              $sphinx_field_name = $index_fields[$field_id]['sphinx_field_name'];

              // Extract sphinx keywords.
              $keywords = $this->extractSphinxKeywords($field_value, $index);

              // Keep only keywords with length > mlt_min_word_len.
              $keywords = array_filter($keywords, $filter_keywords);

              if (!empty($keywords)) {
                // Escape the keywords.
                $keywords = array_map(array('SearchApiSphinxService', 'escapeString'), $keywords);

                // Concatenate with the keyword separator previously defined.
                $keywords = '"' . implode('"' . $keyword_separator . '"', $keywords) . '"';

                // Add the full text query.
                $search_query_full_text[] = "(@{$sphinx_field_name} ({$keywords}))";
              }
            }
          }

          // Create the appropriate filters for the other fields.
          if (!empty($mlt_fields)) {
            foreach ($mlt_fields as $field_id => $field_info) {
              // Make sure it's a valid indexed field.
              if (isset($index_fields[$field_id])) {
                $value = $field_info['value'];

                // If multi-valued field, create a filter with OR conjunction.
                if (is_array($value)) {
                  $filter = new SearchApiQueryFilter('OR');
                  foreach ($value as $sub_value) {
                    $filter->condition($field_id, $sub_value, '=');
                  }
                  $query->filter($filter);
                }
                // Else just add a new condition.
                else {
                  $query->condition($field_id, $value, '=');
                }
              }
            }
          }
        }
      }
    }

    // Extract keys - Full text search.
    $query_full_text_fields = array();
    $query_search_string = '';

    $keys = $query->getKeys();
    if (!empty($keys)) {
      if (is_string($keys)) {
        $keys = array($keys);
      }

      // Full text fields in which to perform the search.
      $query_full_text_fields = $query->getFields();
      $sphinx_full_text_fields = array();
      if (!empty($query_full_text_fields)) {
        foreach ($query_full_text_fields as $field_id) {
          if (isset($index_fields[$field_id])) {
            $sphinx_full_text_fields[] = $index_fields[$field_id]['sphinx_field_name'];
          }
        }
      }

      // Query string
      $query_search_string = $this->flattenKeys($keys, $query_options['parse mode'], $sphinx_full_text_fields);

      // Add the 'all' dummy field to allow exclusion-only search.
      $search_query_full_text[] = '@sphinx_all "' . $this->getSphinxSelectAllKeyword($index) . '"';

      if (!empty($query_search_string)) {
        $field_list = '';

        // Add field restriction.
        if (!empty($sphinx_full_text_fields)) {
          $field_list = '@(' . implode(',', $sphinx_full_text_fields) . ') ';
        }

        $search_query_full_text[] = $field_list . $query_search_string;
      }
    }

    // Parse filters.
    $parsed_query_filters = $this->parseFilter($query->getFilter(), $index_fields, $search_query_expressions);
    $search_query_filters = array_merge($search_query_filters, $parsed_query_filters);

    // Main search query.
    $search_query = array(
      'fields' => $search_query_fields,
      'expressions' => $search_query_expressions,
      'full text' => $search_query_full_text,
      'filters' => $search_query_filters,
      'group by' => $search_query_group_by,
      'order by' => $search_query_sort,
      'limit' => $search_query_limit,
      'options' => $search_query_options,
    );

    // Execute main search query.
    $results = $this->executeSphinxSearchQuery($index_name, $search_query);

    // Results.
    if (isset($results)) {
      $matches = array();

      while (($result = $results->fetchAssoc()) !== FALSE) {
        $id = $result['id'];
        $score = $result['score'];

        // Only keep relevant fields.
        // (Discard id, score and other sphinx specific fields if any).
        $fields = array();
        foreach ($result as $sphinx_field_id => $sphinx_field_value) {
          if (isset($sphinx_fields[$sphinx_field_id])) {
            $field_id = $sphinx_fields[$sphinx_field_id];
            $type = $index_fields[$field_id]['sphinx_field_type'];
            $fields[$field_id] = $this->convertSphinxValueToFieldValue($type, $sphinx_field_value);
          }
        }

        $matches[$id] = array(
          'id' => $id,
          'score' => $score,
          'fields' => $fields,
        );
      }

      $search_result['results'] = $matches;
    }

    // Result count.
    $result_count = $this->getSphinxQueryMeta('total_found');
    if (!isset($result_count)) {
      $result_count = !empty($search_result['results']) ? count($search_result['results']) : 0;
    }
    $search_result['result count'] = $result_count;


    // SEARCH API FACETS.
    if ($facets = $query->getOption('search_api_facets')) {
      // Active search corresponding to this index.
      $searcher = key(facetapi_get_active_searchers());
      // Get the FacetApiAdpater for this searcher.
      $adapter = isset($searcher) ? facetapi_adapter_load($searcher) : NULL;

      // "Missing" values only work for string ('') and MVA (see emptyMVA)
      // because NULL is saved as 0 which is a valid value
      // for timestamp, integer, boolean and so on.
      $missing = array(
        'rt_attr_multi' => $this->emptyMVA,
        'rt_attr_string' => '',
      );

      // Timezone with which daemon was started and used in
      // DateTime::createFromFormat() when handling date
      // related facets.
      $timezone = new DateTimeZone($this->getOption('timezone', 'UTC'));

      // Loop trough facets.
      foreach ($facets as $facet_id => $facet_info) {
        $field_id = $facet_info['field'];

        // Skip if not recognized as a known field.
        if (!isset($index_fields[$field_id])) {
          continue;
        }

        $facet_limit = empty($facet_info['limit']) ? $max_matches : $facet_info['limit'];
        if ($facet_limit > $max_matches || $facet_limit < 0) {
          $facet_limit = $max_matches;
        }
        $facet_min_count = $facet_info['min_count'];
        $facet_missing = $facet_info['missing'];

        $sphinx_field_name = $index_fields[$field_id]['sphinx_field_name'];
        $sphinx_field_type = $index_fields[$field_id]['sphinx_field_type'];

        // OR facet conjunction. We discard filters on the facet field.
        // TODO: Need to distinguish between facet filters and other filters.
        if (isset($facet_info['operator']) && $facet_info['operator'] == 'or') {
          $facet_query_expressions = array();
          $facet_query_filters = $this->parseFilter($query->getFilter(), $index_fields, $facet_query_expressions, $field_id);
        }
        else {
          $facet_query_expressions = $search_query_expressions;
          $facet_query_filters = $search_query_filters;
        }

        $date_format = '';
        $date_gap = 'DAY';

        // Date facet.
        if ($sphinx_field_type === 'rt_attr_timestamp') {
          $gap_weight = array('YEAR' => 2, 'MONTH' => 1, 'DAY' => 0);
          $gaps = array();

          // If we have more results than the maximum of matches then
          // we try to determinate the best gap.
          if ($result_count > $max_matches) {
            // Get the date granularity.
            if (isset($adapter)) {
              // Get the current date gap from the active date filters.
              $active_items = $adapter->getActiveItems(array('name' => $facet_id));
              if (!empty($active_items)) {
                foreach ($active_items as $active_item) {
                  $value = $active_item['value'];
                  if (strpos($value, ' TO ') > 0) {
                    list($date_min, $date_max) = explode(' TO ', str_replace(array('[', ']'), '', $value), 2);
                    $gap = SearchApiSphinxService::getDateGap($date_min, $date_max, FALSE);
                    if (isset($gap_weight[$gap])) {
                      $gaps[] = $gap_weight[$gap];
                    }
                  }
                }
                if (!empty($gaps)) {
                  // Minimum gap.
                  $date_gap = array_search(min($gaps), $gap_weight);
                }
              }
              // Get the min and max dates to determine the initial date gap.
              else {
                $query = sprintf('SELECT %s FROM %s WHERE %s > 0 ORDER BY %s ASC LIMIT 1', $sphinx_field_name, $index_name, $sphinx_field_name, $sphinx_field_name);
                $date_min = $this->querySphinx($query)->fetchField();

                $query = sprintf('SELECT %s FROM %s WHERE %s > 0 ORDER BY %s DESC LIMIT 1', $sphinx_field_name, $index_name, $sphinx_field_name, $sphinx_field_name);
                $date_max = $this->querySphinx($query)->fetchField();

                if (!empty($date_min) && !empty($date_max)) {
                  $date_gap = SearchApiSphinxService::getDateGap($date_min, $date_max);
                }
              }
            }
          }

          switch ($date_gap) {
            // Already a selected YEAR, we want the months.
            case 'YEAR':
              $date_format = 'Ym';
              $facet_query_field = sprintf('YEARMONTH(%s) AS facet_query_field', $sphinx_field_name);
              $facet_query_group = 'facet_query_field';
              break;

            // Already a selected MONTH, we want the days.
            case 'MONTH':
              $date_format = 'Ymd';
              $facet_query_field = sprintf('YEARMONTHDAY(%s) AS facet_query_field', $sphinx_field_name);
              $facet_query_group = 'facet_query_field';
              break;

            // Already a selected DAY, we want the hours and so on.
            case 'DAY':
              $date_format = 'U';
              $facet_query_field = $sphinx_field_name;
              $facet_query_group = $sphinx_field_name;
              break;

            // By default we return result counts by year.
            default:
              $date_format = 'Y';
              $facet_query_field = sprintf('YEAR(%s) AS facet_query_field', $sphinx_field_name);
              $facet_query_group = 'facet_query_field';
          }
        }
        // TODO: range facets?
        else {
          $facet_query_field = $sphinx_field_name;
          $facet_query_group = $sphinx_field_name;
        }

        // Facet query fields.
        $facet_query_fields = array(
          '@groupby',
          $facet_query_field,
          'COUNT(*) AS result_count',
        );

        // Facet query.
        $facet_query = $search_query;
        $facet_query['fields'] = $facet_query_fields;
        $facet_query['expressions'] = $facet_query_expressions;
        $facet_query['filters'] = $facet_query_filters;
        $facet_query['group by'] = array($facet_query_group);
        $facet_query['order by'] = array('result_count' => 'DESC');
        $facet_query['limit'] = array($facet_limit);

        // Execute facet query.
        $results = $this->executeSphinxSearchQuery($index_name, $facet_query);

        if (isset($results)) {
          while (($result = $results->fetchAssoc()) !== FALSE) {
            // If date_format is set then it's a date facet
            // and we convert the date to a unix timestamp.
            if ($date_format !== '') {
              $date = DateTime::createFromFormat('!' . $date_format, $result['@groupby'], $timezone);
              $value = $date->getTimestamp();
            }
            // Use sphinx field name instead of @groupby field for strings
            // because @groupby is a hashed value of the string.
            elseif ($sphinx_field_type === 'rt_attr_string' && isset($result[$sphinx_field_name])) {
              $value = $result[$sphinx_field_name];
            }
            else {
              $value = $result['@groupby'];
            }

            // Handle missing facets.
            if (isset($missing[$sphinx_field_type]) && $value == $missing[$sphinx_field_type]) {
              if (!$facet_missing) {
                continue;
              }
              else {
                $value = '!';
              }
            }

            // Add the facet if result count > facet_min_count.
            if ($result['result_count'] >= $facet_min_count) {
              $search_result['search_api_facets'][$facet_id][] = array(
                'count' => $result['result_count'],
                'filter' => '"' . $value . '"',
              );
            }
          }
        }
      }
    }


    // Post search operations.
    if (!empty($search_result['results'])) {
      // Snippets.
      $excerpt = $this->getIndexOption($index, 'snippets', 'excerpt');
      $highlight = $this->getIndexOption($index, 'snippets', 'highlight');

      $ids = array_keys($search_result['results']);

      // Build excerpts or highlight content.
      $result = $this->buildSphinxExcerpts($ids, $index, $query_full_text_fields, $query_search_string, $excerpt, $highlight);

      if (!empty($result)) {
        foreach ($result as $id => $item) {
          // Highlighted content.
          if (isset($item['fields'])) {
            // Field content is replaced with highlighted one.
            foreach ($item['fields'] as $field_id => $field_value) {
              $search_result['results'][$id]['fields'][$field_id] = $field_value;
            }
          }
          // Excerpt.
          if (isset($item['excerpt'])) {
            $search_result['results'][$id]['excerpt'] = $item['excerpt'];
          }
        }
      }
    }

    return $search_result;
  }

  /**
   * Helper function. Return the meta information from the last executed query.
   */
  protected function getSphinxQueryMeta($key = NULL) {
    // SphinxQL query. See querySphinx notes.
    $result = $this->querySphinx('SHOW META')->fetchAllKeyed();
    if (isset($key)) {
      return isset($result[$key]) ? $result[$key] : NULL;
    }
    else {
      return $result;
    }
  }

  /**
   * Build and execute a sphinx search query and return the result set.
   *
   * Note: all the values passed in the search_query have been previously
   * sanitized to prevent SQL injection.
   *
   * @param array $search_query
   *   Associative array containing the following keys:
   *   - fields: array of select fields (ex: 'id', 'sphinx_name AS name'...).
   *   - expresssions: array of expressions (ex: 'id+1 AS expression_1'...).
   *   - full text: array of full text queries (ex: 'humanitarian|conflict').
   *   - filters: array of filters (WHERE) (ex: 'id=3', 'expression_1<>0').
   *   - group by: array of group by statements (ex: 'sphinx_name').
   *   - order by: array, key as field to sort by and value as DESC or ASC.
   *   - limit: array with 1 or 2 values (LIMIT offset, limit).
   *   - options: array of sphinxQL options. See documentation.
   *   Everything is optional except for 'fields'.
   */
  protected function executeSphinxSearchQuery($index_name, $search_query = NULL) {
    $results = NULL;

    if (!empty($index_name) && !empty($search_query)) {
      $query_parts = array();

      $select_parts = array();

      // Fields - mandatory.
      if (empty($search_query['fields'])) {
        return NULL;
      }
      else {
        $select_parts = array_merge($select_parts, $search_query['fields']);
      }

      // Expressions.
      if (!empty($search_query['expressions'])) {
        $select_parts = array_merge($select_parts, $search_query['expressions']);
      }

      $query_parts[] = sprintf("SELECT %s", implode(',', $select_parts));

      // Index.
      $query_parts[] = sprintf("FROM %s", $index_name);

      // Conditions.
      $conditions = array();

      // Full text queries.
      if (!empty($search_query['full text'])) {
        $conditions[] = sprintf("MATCH('@@relaxed %s')", implode(' ', $search_query['full text']));
      }

      // Filters.
      if (!empty($search_query['filters'])) {
        $conditions = array_merge($conditions, $search_query['filters']);
      }

      if (!empty($conditions)) {
        $query_parts[] = sprintf("WHERE %s", implode(' AND ', $conditions));
      }

      // Group by.
      if (!empty($search_query['group by'])) {
        $query_parts[] = sprintf("GROUP BY %s", implode(',', $search_query['group by']));
      }

      // Order by.
      if (!empty($search_query['order by'])) {
        $order_by = array();
        foreach ($search_query['order by'] as $column => $direction) {
          $order_by[] = sprintf("%s %s", $column, $direction);
        }
        $query_parts[] = sprintf("ORDER BY %s", implode(',', $order_by));
      }

      // Limit.
      if (!empty($search_query['limit'])) {
        if (count($search_query['limit']) > 1) {
          $query_parts[] = sprintf("LIMIT %u,%u", $search_query['limit'][0], $search_query['limit'][1]);
        }
        else {
          $query_parts[] = sprintf("LIMIT %u", $search_query['limit'][0]);
        }
      }

      // Options.
      if (!empty($search_query['options'])) {
        $options = array();
        foreach ($search_query['options'] as $option => $value) {
          $options[] = sprintf('%s=%s', $option, $value);
        }
        $query_parts[] = sprintf("OPTION %s", implode(',', $options));
      }

      if (!empty($query_parts)) {
        $query = implode(' ', $query_parts);
        // SphinxQL query. See querySphinx notes.
        $results = $this->querySphinx($query);
      }
    }

    return $results;
  }

  /**
   * Return a sphinx full text search query.
   *
   * Note: values are sanitized using escapeString().
   */
  protected function flattenKeys($keys, $parse_mode = '', $full_text_fields = array()) {
    $conjunction = isset($keys['#conjunction']) ? $keys['#conjunction'] : 'AND';
    $negation = !empty($keys['#negation']);
    $values = array();

    foreach (element_children($keys) as $key) {
      $value = $keys[$key];

      if (empty($value)) {
        continue;
      }

      if (is_array($value)) {
        $values[] = $this->flattenKeys($value);
      }
      elseif (is_string($value)) {
        // If parse mode is not "direct": quote the keyword and escape dashes.
        if ($parse_mode !== 'direct') {
          $value = '"' . str_replace('-', '\\-', $value) . '"';
        }

        // Escape string to prevent SQL injection.
        $value = SearchApiSphinxService::escapeString($value);

        // Unescape full text fields.
        if ($parse_mode === 'direct' && !empty($full_text_fields)) {
          $field_replacements = array();
          foreach ($full_text_fields as $field) {
            $field_replacements['\\@' . $field] = '@' . $field;
          }
          $value = strtr($value, $field_replacements);
        }


        $values[] = $value;
      }
    }

    if (!empty($values)) {
      return ($negation === TRUE ? '!' : '') . '(' . implode($conjunction === 'AND' ? ' ' : '|', $values) . ')';
    }
    else {
      return '';
    }
  }

  /**
   * Return a list of sphinx filters and update the list of expressions.
   *
   * Note: values are sanitized using convertFieldValueToSphinxValue().
   * We can not use placeholders because PDO doesn't handle floats
   * or unsigned integers and current Drupal usage of PDO results
   * in all parameters to be quoted.
   *
   * See https://bugs.php.net/bug.php?id=44639
   * See http://php.net/manual/en/pdo.constants.php
   */
  protected function parseFilter(SearchApiQueryFilter $query_filter, $index_fields, &$expressions, $ignored_field_id = '') {
    if (empty($query_filter)) {
      return NULL;
    }
    else {
      $filters = array();
      foreach ($query_filter->getFilters() as $filter_info) {
        // Simple filter [field_id, value, operator].
        if (is_array($filter_info)) {
          $field_id = $filter_info[0];

          // If field should be ignored, we skip.
          if ($field_id === $ignored_field_id) {
            continue;
          }
          elseif (isset($index_fields[$field_id])) {
            $sphinx_field_name = $index_fields[$field_id]['sphinx_field_name'];
            $sphinx_field_type = $index_fields[$field_id]['sphinx_field_type'];

            // Sanitize the value to prevent SQL injection.
            $filter_value = $this->convertFieldValueToSphinxValue($sphinx_field_type, $filter_info[1]);
            $filter_operator = str_replace('!=', '<>', $filter_info[2]);

            // Check filter value validity.
            if (is_null($filter_value)) {
              throw new SearchApiException(t("Invalid filter value '!filter_value'. Field: '!field_id'.", array(
                '!filter_value' => $filter_info[1],
                '!field_id' => $field_id,
              )));
            }

            // Check filter operator validity.
            if (!SearchApiSphinxService::checkFilterOperator($sphinx_field_type, $filter_operator)) {
              throw new SearchApiException(t("Invalid filter operator '!filter_operator'. Field: '!field_id'.", array(
                '!filter_operator' => $filter_operator,
                '!field_id' => $field_id,
              )));
            }

            switch ($sphinx_field_type) {
              // In case of string attributes, we compare crc32 values.
              // We need to use the non-converted value.
              case 'rt_attr_string':
                $filter_value = empty($filter_info[1]) ? '' : $filter_info[1];
                $value = sprintf("%u", crc32($filter_value));
                $alias = 'expression_' . count($expressions);
                $expressions[] = "CRC32({$sphinx_field_name}) AS {$alias}";
                $filters[] = "{$alias}{$filter_operator}{$value}";
                break;

              // Operator is one of =, <=, >=, <, > or <>.
              case 'rt_attr_uint':
              case 'rt_attr_timestamp':
              case 'rt_attr_float':
                $filters[] = "{$sphinx_field_name}{$filter_operator}{$filter_value}";
                break;

              // We build an expression, so it can be used by parent filters.
              case 'rt_attr_multi':
                $filter_value = implode(',', is_array($filter_value) ? $filter_value : array($filter_value));
                $alias = 'expression_' . count($expressions);
                $expressions[] = "IN({$sphinx_field_name},{$filter_value}) AS {$alias}";
                $filters[] = "{$alias}{$filter_operator}1";
                break;

              // Field type is not valid for filtering.
              case 'rt_field':
              default:
                throw new SearchApiException(t("Invalid filter type '!field_type' for filtering. Field: '!field_id'.", array(
                  '!field_type' => $index_fields[$field_id]['type'],
                  '!field_id' => $field_id,
                )));
            }
          }
          // Invalid filter.
          else {
            throw new SearchApiException(t("Invalid filter on unknown or unindexed field '!field_id'", array(
              '!field_id' => $field_id,
            )));
          }
        }
        // Nested filters.
        elseif ($filter_info instanceof SearchApiQueryFilter) {
          $nested_filters = $this->parseFilter($filter_info, $index_fields, $expressions, $ignored_field_id);
          if (!empty($nested_filters)) {
            $filters = array_merge($filters, $nested_filters);
          }
        }
      }

      // In case of OR conjunction, we build an expression with the filters.
      if ($query_filter->getConjunction() === 'OR' && count($filters) > 1) {
        $alias = 'expression_' . count($expressions);
        $expression = '(' . implode(') OR (', $filters) . ')';
        $expressions[] = "({$expression}) AS {$alias}";
        $filters = array("{$alias}=1");
      }

      return $filters;
    }
  }


  /**
   * Build excerpt and/or highlight keywords in content for a list of entities.
   *
   * @param array $ids
   *   Array of entity ids.
   *
   * @param SearchApiIndex $index
   *   Search API index.
   *
   * @param array $fields
   *   Array of fields used to get the snippets.
   *
   * @param string $query
   *   Sphinx full text search query.
   *
   * @param boolean $excerpt
   *   Whether or not build excerpts
   *
   * @param boolean $highlight
   *   Whether or not highlight searched keywords.
   */
  protected function buildSphinxExcerpts(array $ids, SearchApiIndex $index, array $fields, $query, $excerpt = FALSE, $highlight = FALSE) {
    if (empty($ids) || empty($fields) || empty($query) || (empty($excerpt) && empty($highlight))) {
      return array();
    }

    $result = array();

    $index_name = $this->getSphinxIndexName($index);

    $index_fields = $index->getFields();

    $full_text_fields = $index->getFulltextFields();

    // Fields used to get the snippets.
    $excerpt_fields = array();
    foreach ($fields as $field_id) {
      if (in_array($field_id, $full_text_fields)) {
        $excerpt_fields[$field_id] = $index_fields[$field_id];
      }
    }

    // Continue only if there are valid fields.
    if (empty($excerpt_fields)) {
      return array();
    }

    $entity_type = $index->item_type;

    // We have to load the entities to get the content.
    $entities = entity_load($entity_type, $ids);

    if (!empty($entities)) {
      foreach ($entities as $entity_id => $entity) {
        // Get the entity wrapper so we can extract the fields and their value.
        $entity_wrapper = entity_metadata_wrapper($entity_type, $entity);

        if (isset($entity_wrapper)) {
          // Get the full fields with their value.
          $excerpt_fields = search_api_extract_fields($entity_wrapper, $excerpt_fields);

          if (!empty($excerpt_fields)) {
            // We extract the field ids.
            $excerpt_field_ids = array_keys($excerpt_fields);

            // Get the content of the fields.
            $texts = array();
            foreach ($excerpt_fields as $field_id => $field_info) {
              $texts[] = $field_info['value'];
            }

            // Get excerpts.
            if (!empty($excerpt)) {
              $options = array(
                'before_match' => "'<em>'",
                'after_match' => "'</em>'",
                'html_strip_mode' => "'strip'",
                'limit' => '300',
              );

              $snippets = $this->extractSphinxSnippets($texts, $index, $query, $options);

              $result[$entity_id]['excerpt'] = implode(' ... ', $snippets);
            }

            // Get highlighted content. It can replace the original content.
            if (!empty($highlight)) {
              $options = array(
                'before_match' => "'<em>'",
                'after_match' => "'</em>'",
                'html_strip_mode' => "'retain'",
                'limit' => '0',
              );

              $snippets = $this->extractSphinxSnippets($texts, $index, $query, $options);

              foreach ($excerpt_field_ids as $key => $field_id) {
                $result[$entity_id]['fields'][$field_id] = $snippets[$key];
              }
            }
          }
        }
      }
    }

    return $result;
  }

  /**
   * Extract snippets from the given texts.
   *
   * It can be used to get excerpts or highlight keywords.
   *
   * @param array $texts
   *   Array of texts from which to get snippets.
   *
   * @param SearchApiIndex $index
   *   Index used to get the settings for snippets generation.
   *
   * @param string $query
   *   String containing the keywords used for building the snippet.
   *
   * @param array $options
   *   Array of options as described in the Sphinx documentation.
   *
   * @see http://sphinxsearch.com/docs/current.html#api-func-buildexcerpts
   * @see http://sphinxsearch.com/docs/current.html#sphinxql-call-snippets
   */
  public function extractSphinxSnippets(array $texts, SearchApiIndex $index, $query, array $options = array()) {
    $index_name = $this->getSphinxIndexName($index);

    // Prepare the texts from which to get snippets.
    $snippets_texts = array();
    foreach ($texts as $text) {
      $snippets_texts[] = "'" . SearchApiSphinxService::escapeString($text) . "'";
    }
    $snippets_texts = implode(',', $snippets_texts);

    // Prepare the options.
    $snippets_options = array();
    foreach ($options as $option => $value) {
      $snippets_options[] = "{$value} AS {$option}";
    }
    $snippets_options = implode(',', $snippets_options);

    $excerpt_query = "CALL SNIPPETS(({$snippets_texts}), '{$index_name}', '{$query}', {$snippets_options})";

    // Execute the query and return the results.
    return $this->querySphinx($excerpt_query)->fetchCol();
  }

  /**
   * Extract sphinx keywords from a text according to the index settings.
   *
   * @see http://sphinxsearch.com/docs/current.html#sphinxql-call-keywords
   */
  public function extractSphinxKeywords($text, SearchApiIndex $index) {
    $index_name = $this->getSphinxIndexName($index);

    // Escape the field value to prevent SQL injection.
    $keywords_text = SearchApiSphinxService::escapeString($text);

    // Extract the keywords in accordance to the index settings.
    $keywords_query = "CALL KEYWORDS('{$keywords_text}', '{$index_name}')";

    // SphinxQL query. See querySphinx notes.
    $keywords = $this->querySphinx($keywords_query)->fetchCol(1);

    // Return normalized keywords.
    return $keywords;
  }

  /**
   * Helper function. Check filter operator's validity for a given field type.
   */
  protected static function checkFilterOperator($type, $operator) {
    switch ($type) {
      case 'rt_attr_string':
      case 'rt_attr_multi':
        return in_array($operator, array('=', '<>'));

      case 'rt_attr_uint':
      case 'rt_attr_timestamp':
      case 'rt_attr_float':
        return in_array($operator, array('=', '<=', '>=', '<', '>', '<>'));
    }

    return FALSE;
  }

  /**
   * Helper function. sanitize list of IDs.
   */
  protected static function sanitizeIDs(array $ids = array()) {
    $result = array();
    foreach ($ids as $id) {
      if (is_numeric($id)) {
        $id = $id + 0;
        if (is_int($id) && $id > 0) {
          $result[] = sprintf('%u', $id);
        }
      }
    }
    return $result;
  }

  /**
   * Helper function. Make sure ID is an integer and prepare it for queries.
   */
  protected static function sanitizeID($id) {
    if (is_numeric($id)) {
      $id = $id + 0;
      if (is_int($id) && $id > 0) {
        return sprintf('%u', $id);
      }
    }
    return NULL;
  }

  /**
   * Helper function. Return date gap from two dates or timestamps.
   *
   * @see facetapi_get_timestamp_gap()
   */
  protected static function getDateGap($min, $max, $timestamp = TRUE) {
    if ($timestamp !== TRUE) {
      $min = strtotime($min);
      $max = strtotime($max);
    }

    if (empty($min) || empty($max)) {
      return 'DAY';
    }

    $diff = $max - $min;

    switch (TRUE) {
      case ($diff > 86400 * 365):
        return 'NONE';

      case ($diff > 86400 * gmdate('t', $min)):
        return 'YEAR';

      case ($diff > 86400):
        return 'MONTH';

      default:
        return 'DAY';
    }
  }

  /**
   * Return an array of sphinx sort criteria in order of importance.
   *
   * Keys are sphinx field ids, values are directions.
   */
  protected function getSphinxQueryOrderBy($query_sort, $index_fields) {
    $order_by = array();

    if (!empty($query_sort)) {
      foreach ($query_sort as $field_id => $direction) {
        // Sort by relevance. Column name is 'score', see search() function.
        if ($field_id === 'search_api_relevance') {
          $order_by['score'] = $direction;
        }
        elseif (isset($index_fields[$field_id])) {
          $sphinx_field_name = $index_fields[$field_id]['sphinx_field_name'];
          $sphinx_field_type = $index_fields[$field_id]['sphinx_field_type'];

          switch ($sphinx_field_type) {
            case 'rt_attr_string':
            case 'rt_attr_uint':
            case 'rt_attr_timestamp':
            case 'rt_attr_float':
            case 'rt_attr_multi':
              $order_by[$sphinx_field_name] = $direction;
              break;

            case 'rt_field':
            default:
              throw new SearchApiException(t("Invalid field type '!field_type' for sorting. Field '!field_id'.", array(
                '!field_type' => $index_fields[$field_id]['type'],
                '!field_id' => $field_id,
              )));
          }
        }
        else {
          throw new SearchApiException(t("Invalid sort criteria on unknown or unindexed field '!field_id'", array('!field_id' => $field_id)));
        }
      }
    }

    return $order_by;
  }

  /**
   * Helper function. Get indexes.
   */
  public function getIndexes() {
    return search_api_index_load_multiple(FALSE, array('server' => $this->server->machine_name));
  }

  /**
   * Helper function. Get (create) the connection to the sphinx daemon.
   */
  protected function getSphinxConnection() {
    $connection_name = $this->getSphinxConnectionName();

    $database = array(
      'database' => '',
      'username' => '',
      'password' => '',
      'host' => $this->getOption('host'),
      'port' => $this->getOption('port'),
      'driver' => 'mysql',
      'prefix' => '',
    );

    // Create the connection if it doesn't already exist.
    Database::addConnectionInfo($connection_name, 'default', $database);

    // Return the connection name.
    return $connection_name;
  }

  /**
   * Helper function. Close connection to sphinx daemon.
   */
  protected function removeSphinxConnection() {
    Database::removeConnection($this->getSphinxConnectionName());
  }

  /**
   * Helper function. Check sphinx connection.
   */
  public function checkSphinxConnection() {
    try {
      // SphinxQL query. See querySphinx notes.
      return $this->querySphinx('SHOW STATUS')->rowCount() > 0;
    }
    catch (Exception $exception) {
      return FALSE;
    }
  }

  /**
   * Helper function. Get the sphinx connection name.
   */
  protected function getSphinxConnectionName() {
    return 'db_connection_' . $this->getSphinxServerName();
  }

  /**
   * Helper function. Execute a query against the sphinx server.
   *
   * Sphinx emulates the mysql protocol and uses its own syntax: SphinxQL.
   *
   * Important note: queries passed through this function use
   * the SphinxQL syntax which doesn't allow to have escaped values
   * other than strings so we can not use standard placeholders as
   * used in DatabaseConnection::query(). Drupal uses PDOStatement::execute()
   * or this function treats all arguments as string as explained at
   * http://us.php.net/manual/en/pdostatement.execute.php resulting
   * in quoted values.
   *
   * Also PDO doesn't handle unsigned integers which are important
   * because we use crc32 for string comparison.
   *
   * This function receives SphinxQL queries containing values that
   * have been sanitized using convertFieldValueToSphinxValue(),
   * sanitizeIDs() or escapeString() so the execution is safe.
   *
   * @see DatabaseConnection::query()
   *
   * @param string $query
   *   SphinxQL query.
   */
  protected function querySphinx($query) {
    return Database::getConnection('default', $this->getSphinxConnection())->query($query);
  }

  /**
   * Helper function. Get sphinx server name.
   */
  protected function getSphinxServerName() {
    global $databases;

    $site_database = $databases['default']['default']['database'];

    return 'sphinx_server_' . $site_database . '_' . $this->server->machine_name;
  }

  /**
   * Helper function. Get sphinx index name.
   *
   * Index is either an index object or a machine name.
   */
  public function getSphinxIndexName($index) {
    global $databases;

    $site_database = $databases['default']['default']['database'];

    $index_machine_name = is_string($index) ? $index : $index->machine_name;

    return SearchApiSphinxService::escapeName('sphinx_index_' . $site_database . '_' . $index_machine_name . '_rt');
  }

  /**
   * Helper function. Get compatible sphinx fields.
   */
  protected function getSphinxIndexFields(SearchApiIndex $index) {
    $fields = array();
    foreach ($index->getFields() as $field_id => $field_info) {
      $sphinx_field_type = $this->getSphinxFieldType($field_info['type'], $field_id);
      if ($sphinx_field_type !== NULL) {
        $field_info['sphinx_field_name'] = SearchApiSphinxService::escapeName('sphinx_' . strtr($field_id, array(':' => '___')));
        $field_info['sphinx_field_type'] = $sphinx_field_type;
        $field_info['inner_type'] = search_api_extract_inner_type($field_info['type']);
        $fields[$field_id] = $field_info;
      }
    }
    return $fields;
  }

  /**
   * Helper function. Convert a field value to a compatible sphinx value.
   *
   * Values are validated and escaped if necessary to prevent SQL injection.
   *
   * Note: it's not possible to store NULL values so they are converted to:
   * - '' (empty string) for string attributes.
   * - 0 for integer, timestamp and boolean.
   * - $emptyMVA for Multi-valued attributes.
   *
   * Unwanted macthes can be returned due to this limitation.
   */
  protected function convertFieldValueToSphinxValue($type, $value) {
    switch ($type) {
      case 'rt_field':
      case 'rt_attr_string':
        if (!isset($value)) {
          return "''";
        }
        elseif (is_array($value)) {
          // If the tokenizer preprocessor has been selected then value
          // is an array of terms. We concatenate (and ignore the score).
          $values = array();
          foreach ($value as $item) {
            $values[] = $item['value'];
          }
          return "'" . SearchApiSphinxService::escapeString(implode(' ', $values)) . "'";
        }
        elseif (is_string($value)) {
          return "'" . SearchApiSphinxService::escapeString($value) . "'";
        }
        break;

      case 'rt_attr_uint':
      case 'rt_attr_timestamp':
      case 'rt_attr_float':
        if (!isset($value)) {
          return '0';
        }
        elseif (is_bool($value)) {
          return $value === TRUE ? '1' : '0';
        }
        elseif (is_numeric($value)) {
          $value = $value + 0;
          if (is_int($value)) {
            return sprintf('%u', $value);
          }
          elseif (is_float($value) && $type === 'rt_attr_float') {
            return sprintf('%f', $value);
          }
        }
        break;

      case 'rt_attr_multi':
        if (!isset($value)) {
          // We need to store a dummy "empty" value to use in expressions.
          // We don't use 0 as it's a valid value for list<integer> fields.
          return array(sprintf('%u', $this->emptyMVA));
        }
        else {
          $value = is_array($value) ? $value : array($value);
          // Check and convert values.
          foreach ($value as $index => $val) {
            if (!is_numeric($val) || !is_int($val + 0)) {
              break 2;
            }
            $value[$index] = sprintf('%u', $val);
          }
          return $value;
        }
    }

    return NULL;
  }

  /**
   * Helper function. Convert a sphinx value to a field value.
   */
  protected function convertSphinxValueToFieldValue($type, $value) {
    switch ($type) {
      case 'rt_attr_string':
        return $value === '' ? '' : $value;

      case 'rt_attr_uint':
      case 'rt_attr_timestamp':
        return empty($value) ? 0 : SearchApiSphinxService::convertToSigned($value);

      case 'rt_attr_float':
        return empty($value) ? 0.0 : $value + 0;

      case 'rt_attr_multi':
        return empty($value) ? array() : array_map(array('SearchApiSphinxService', 'convertToSigned'), explode(',', $value));

      default:
        return NULL;
    }
  }

  /**
   * Helper function. Convert an unsigned integer to a signed integer.
   */
  public static function convertToSigned($integer) {
    return 0xFFFFFFFF & ($integer + 0);
  }

  /**
   * Helper function. Escape a field or index name.
   *
   * Force names to be strictly alphanumeric-plus-underscore.
   */
  public static function escapeName($name) {
    return preg_replace('/[^A-Za-z0-9_]+/', '', $name);
  }

  /**
   * Helper function. Escape sphinx special characters.
   *
   * @see http://sphinxsearch.googlecode.com/svn/trunk/api/sphinxapi.php
   */
  public static function escapeString($string) {
    $from = array(
      '\\', "\0", "\n", "\r", "'", '"', "\x1a",
      '(', ')', '|', '-', '!', '@',
      '~', '&', '/', '^', '$', '=',
    );

    $to = array(
      '\\\\', '\\0', '\\n', '\\r', "\\'", '\\"', '\\Z',
      '\\(', '\\)', '\\|', '\\-', '\\!', '\\@',
      '\\~', '\\&', '\\/', '\\^', '\\$', '\\=',
    );

    return str_replace($from, $to, $string);
  }

  /**
   * Helper function. Unescape sphinx special characters.
   */
  public static function unescapeString($string) {
    $from = array(
      "\\\\", "\\0", "\\n", "\\r", "\Z", "\'", '\"',
      '\(', '\)', '\|', '\-', '\!', '\@',
      '\~', '\&', '\/', '\^', '\$', '\=',
    );

    $to = array(
      "\\", "\0", "\n", "\r", "\x1a", "'", '"',
      '(', ')', '|', '-', '!', '@',
      '~', '&', '/', '^', '$', '=',
    );

    return str_replace($from, $to, $string);
  }

  /**
   * Helper function. Get the sphinx field type matching the given field type.
   *
   * @param string $type
   *   Field type as returned by search_api_default_field_types().
   *
   * @param string $field
   *   Field name.
   *
   * @return string
   *   Sphinx field type or NULL if none matched the given field type.
   */
  public function getSphinxFieldType($type, $field = '') {
    switch ($type) {
      case 'text':
        return 'rt_field';

      case 'uri':
      case 'string':
      case 'token':
        return 'rt_attr_string';

      case 'integer':
      case 'duration':
      case 'boolean':
        return 'rt_attr_uint';

      case 'decimal':
        return 'rt_attr_float';

      case 'date':
        return 'rt_attr_timestamp';

      case 'list<integer>':
        return 'rt_attr_multi';

      default:
        watchdog('search_api_sphinx', "'Non recognized field type '@type' for field '@field'", array(
          '@type' => $type,
          '@field' => $field,
        ));
        return NULL;
    }
  }

  /**
   * Helper function. Return server options.
   */
  public function getOptions() {
    return $this->options;
  }

  /**
   * Helper function. Return a server option.
   */
  public function getOption($option, $default = NULL) {
    $options = $this->getOptions();
    return isset($options[$option]) ? $options[$option] : $default;
  }

  /**
   * Helper function. Return index sphinx options.
   */
  public function getIndexOptions(SearchApiIndex $index, $category = NULL) {
    if (isset($index->options['sphinx options'])) {
      $options = $index->options['sphinx options'];
      if (isset($category)) {
        return isset($options[$category]) ? $options[$category] : array();
      }
      return $options;
    }
    return array();
  }

  /**
   * Helper function. Return an index sphinx option.
   */
  public function getIndexOption(SearchApiIndex $index, $category, $option, $default = NULL) {
    $options = $this->getIndexOptions($index, $category);
    return isset($options[$option]) ? $options[$option] : $default;
  }

  /**
   * Helper function. Get a "select all" keyword depending on "min_word_len".
   */
  public function getSphinxSelectAllKeyword(SearchApiIndex $index) {
    $min_word_len = $this->getIndexOption($index, 'index', 'min_word_len', 1);
    return $min_word_len > 3 ? str_pad('all', 'l', $min_word_len) : 'all';
  }

  /**
   * Helper function. Start/stop sphinx (searchd) daemon.
   *
   * @param string $command
   *   Either 'start' or 'stop'.
   */
  public function startStopSphinxDaemon($command = 'start') {
    $searchd = $this->getOption('searchd');

    $timezone = $this->getOption('timezone', 'UTC');

    $configuration_file = $this->getSphinxServerConfigurationFile();

    if (!file_exists($configuration_file)) {
      watchdog('search_api_sphinx_daemon', "The configuration file doesn't exist. The daemon was not started.");
      return FALSE;
    }

    // Only set the timezone if not running under windows.
    if (drupal_strtoupper(drupal_substr(PHP_OS, 0, 3)) === 'WIN') {
      if ($command === 'stop') {
        $searchd_command = "net stop SphinxSearch && {$searchd} --delete --config \"{$configuration_file}\" --servicename SphinxSearch && sc start SphinxSearch";
      }
      else {
        $searchd_command = "{$searchd} --install --config \"{$configuration_file}\" --servicename SphinxSearch && net start SphinxSearch";
      }
    }
    else {
      // We wait for 1 second after the 'start' command
      // because there is a tiny delay (milliseconds)
      // before the server can be queried.
      $searchd_command = "TZ=\"{$timezone}\" {$searchd} --config \"{$configuration_file}\"" . ($command === 'stop' ? ' --stopwait' : ';sleep 1');
    }

    $output = array();

    exec($searchd_command, $output, $return_val);

    watchdog('search_api_sphinx_daemon', "@action\nCommand: @command\nReturn value: @return_val\n\nOutput: @output", array(
      '@action' => drupal_strtoupper($command),
      '@command' => $searchd_command,
      '@return_val' => $return_val,
      '@output' => implode("\n", (array) $output),
    ));

    // If start, we check the connection.
    if ($command === 'start') {
      return $this->checkSphinxConnection();
    }

    return TRUE;
  }

  /**
   * Helper function. Get sphinx server configuration file.
   */
  protected function getSphinxServerConfigurationFile() {
    return $this->getSphinxServerDirectory() . '/sphinx.conf';
  }

  /**
   * Helper function. Delete index files.
   *
   * @param SearchApiIndex|array|string $index
   *   An index object, an array of indexes or a machine name.
   */
  protected function deleteSphinxIndexFiles($index = NULL) {
    if (!isset($index)) {
      return;
    }
    $indexes = is_array($index) ? $index : array($index);

    $server_directory = $this->getSphinxServerDirectory();

    foreach ($indexes as $index) {
      $files = new GlobIterator($server_directory . '/' . $this->getSphinxIndexName($index) . '.*');

      foreach ($files as $file) {
        unlink($file->getPathname());
      }
    }
  }

  /**
   * Helper function. Update (write/remove) sphinx server configuration file.
   *
   * @param string $extra_configuration
   *   Extra configuration to be added to the server configuration.
   *   For fast indexing for example.
   */
  protected function updateSphinxServerConfigurationFile($extra_configuration = NULL) {
    $server_configuration_file = $this->getSphinxServerConfigurationFile();

    // Remove the existing configuration file.
    @unlink($server_configuration_file);

    // Recreate the server configuration file.
    $configuration = array();

    // Get configuration of all indexes on this server.
    foreach ($this->getIndexes() as $index) {
      $configuration[] = $this->getSphinxIndexConfiguration($index);
    }

    // Only write the file if there are indexes to serve.
    if (!empty($configuration)) {
      if (!empty($extra_configuration)) {
        $configuration[] = $extra_configuration;
      }

      $configuration[] = $this->getSphinxServerConfiguration();

      // Merge indexes and server configurations.
      file_put_contents($server_configuration_file, implode("\n", $configuration));

      // File was successfully created.
      return TRUE;
    }

    // No server configuration file.
    return FALSE;
  }

  /**
   * Helper function. Get sphinx index configuration.
   */
  protected function getSphinxIndexConfiguration(SearchApiIndex $index) {
    $configuration = array();

    $index_name = $this->getSphinxIndexName($index);

    // Index options.
    $index_options = array(
      'type' => 'rt',
      'path' => $this->getSphinxServerDirectory() . '/' . $index_name,
      'dict' => 'keywords',
    );

    $index_options += $this->getIndexOptions($index, 'index');

    foreach ($index_options as $option => $value) {
      $value = str_replace(array("\r", "\n"), array('', "\\\n\t"), $value);
      $configuration[] = "{$option} = {$value}";
    }

    // Fields.
    $fields = $this->getSphinxIndexFields($index);

    foreach ($fields as $field_id => $field_info) {
      $configuration[] = "{$field_info['sphinx_field_type']} = {$field_info['sphinx_field_name']}";
    }

    // Dummy field, used to select all documents (@sphinx_all).
    $configuration[] = 'rt_field = sphinx_all';


    // Index configuration.
    return "index {$index_name}\n{\n\t" . implode("\n\t", $configuration) . "\n}\n";
  }

  /**
   * Helper function. Get sphinx server configuration.
   */
  protected function getSphinxServerConfiguration() {
    $configuration = array();

    $server_name = $this->getSphinxServerName();

    $directory = $this->getSphinxServerDirectory();

    // Server options.
    $server_options = array(
      'listen'                  => $this->getOption('host') . ':' . $this->getOption('port') . ':mysql41',
      'pid_file'                => $directory . '/searchd.pid',
      'log'                     => $directory . '/searchd.log',
      'query_log'               => $directory . '/query.log',
      'binlog_path'             => $directory,
      'max_children'            => $this->getOption('max_children'),
      'max_matches'             => $this->getOption('max_matches'),
      'read_timeout '           => '5',
      'seamless_rotate'         => '1',
      'preopen_indexes'         => '1',
      'unlink_old'              => '1',
      'workers'                 => 'threads',
      'compat_sphinxql_magics'  => '0',
      'rt_flush_period'         => $this->getOption('rt_flush_period'),
      'expansion_limit'         => $this->getOption('expansion_limit'),
      'collation_server'        => $this->getOption('collation_server'),

    );

    foreach ($server_options as $option => $value) {
      $configuration[] = "{$option} = {$value}";
    }

    // Server configuration.
    return "searchd\n{\n\t" . implode("\n\t", $configuration) . "\n}\n";
  }

  /**
   * Helper function. Create the server directory.
   *
   * @throws SearchApiException
   *   If the server directory could not be created.
   */
  protected function createSphinxServerDirectory($directory = NULL) {
    if (!isset($directory)) {
      $directory = $this->getSphinxServerDirectory();
    }

    if (!mkdir($directory, 0700)) {
      throw new SearchApiException(t("The directory '!directory' is not accessible or can not be created.", array(
        '!directory' => $directory,
      )));
    }
  }

  /**
   * Helper function. Delete the server directory and its content.
   */
  protected function deleteSphinxServerDirectory($directory = NULL) {
    if (!isset($directory)) {
      $directory = $this->getSphinxServerDirectory();
    }

    if (!is_dir($directory)) {
      return;
    }

    $iterator = new RecursiveIteratorIterator(
      new RecursiveDirectoryIterator($directory),
      RecursiveIteratorIterator::CHILD_FIRST
    );

    foreach ($iterator as $file) {
      if ($file->isDir()) {
        rmdir($file->getPathname());
      }
      elseif ($file->isFile() || $file->isLink()) {
        unlink($file->getPathname());
      }
    }

    rmdir($directory);
  }

  /**
   * Helper function. Update the server directory if necessary.
   *
   * Before calling this function, the daemon has to be stopped.
   */
  protected function updateSphinxServerDirectory() {
    if (empty($this->server->original->options['base_directory'])) {
      return FALSE;
    }

    $current_directory = $this->getSphinxServerDirectory();
    $previous_directory = $this->getSphinxServerDirectory($this->server->original->options['base_directory']);

    if ($current_directory !== $previous_directory) {
      // We delete the previous directory and its content.
      $this->deleteSphinxServerDirectory($previous_directory);

      // And we create the new one.
      $this->createSphinxServerDirectory($current_directory);

      // The directory has been updated.
      return TRUE;
    }

    return FALSE;
  }

  /**
   * Helper function. Return the directory for this server.
   */
  protected function getSphinxServerDirectory($base_directory = NULL) {
    if (!isset($base_directory)) {
      $base_directory = $this->getOption('base_directory');
    }
    return $base_directory . '/' . $this->getSphinxServerName();
  }

  /**
   * Helper function. Update the sphinx server.
   *
   * - stop the daemon.
   * - delete the index files if index is provided.
   * - update the server directory if necessary.
   * - rebuild the configuration file.
   * - restart the daemon.
   *
   * @param object|array|string $index
   *   An index object, an array of indexes or a machine name.
   *
   * @throws SearchApiException
   *   If restarting the sphinx daemon failed.
   */
  public function updateSphinxServer($index = NULL) {
    // First we stop the daemon.
    // All indexes for this server are flushed to disk.
    $this->startStopSphinxDaemon('stop');

    // Remove connection to sphinx daemon.
    $this->removeSphinxConnection();

    // Then we delete the index files for the given index(es).
    $this->deleteSphinxIndexFiles($index);

    // Update the server directory if necessary.
    $updated = $this->updateSphinxServerDirectory();

    // We recreate the server configuration file.
    $rebuilt = $this->updateSphinxServerConfigurationFile();

    // And finally we restart the daemon if configuration was written.
    if ($rebuilt && !$this->startStopSphinxDaemon('start')) {
      throw new SearchApiException(t('Error while starting the sphinx daemon. Please read the log for more information.'));
    }

    // Flag for re-indexing.
    return $rebuilt && $updated;
  }

  /**
   * Create a Sphinx disk Index and attach it to an empty RT index.
   *
   * This a really specific function that has to be called from a drush
   * function and that supports limited index options.
   *
   * Indexing this way is much faster but not all Search API options,
   * processors and so on are supported.
   * Especially processors will no be used resulting in a different
   * result from normal indexing.
   * In case indexing can not be done, this function returns FALSE.
   *
   * Suppports Aggregation Fields for fields that are normally indexed.
   * However it doesn't support the COUNT type aggregation.
   *
   * Supports bundles.
   *
   * Doesn't support fields of related entities. For example, if the index
   * item type is 'node', the email field of the associated user can not
   * be indexed. It only supports Entity schema fields and
   * associated fields (Field API).
   */
  public function fastIndexing($index) {
    // Not called from drush.
    if (!function_exists('drush_get_context')) {
      return FALSE;
    }

    // Get index name.
    $index_name = $this->getSphinxIndexName($index);

    // Get index fields.
    $index_fields = $this->getSphinxIndexFields($index);

    // Entity type.
    $entity_type = $index->item_type;

    // Get entity info for the item type the index works on.
    $entity_info = entity_get_info($entity_type);
    $entity_property_info = entity_get_property_info($entity_type);
    $entity_properties = $entity_property_info['properties'];

    $entity_table = $entity_info['base table'];
    $entity_id_field = $entity_info['entity keys']['id'];

    $entity_schema = drupal_get_schema($entity_table);

    $query_main_joins = array();
    $query_main_fields = array();
    $query_attr_simple = array();
    $query_attr_multi = array();

    // Sphinx default fields.
    $query_main_fields[] = "{$entity_table}.{$entity_id_field} AS id";
    $query_main_fields[] = sprintf("'%s' AS sphinx_all", $this->getSphinxSelectAllKeyword($index));

    // Start and end used in ranged queries.
    $start = '$start';
    $end = '$end';

    // Bundle handling.
    $entity_bundle_field = '';
    $bundle_condition = '';
    if (!empty($index->options['data_alter_callbacks']['search_api_alter_bundle_filter']['settings']['bundles'])) {
      $search_api_bundle_settings = $index->options['data_alter_callbacks']['search_api_alter_bundle_filter']['settings'];

      // If a bundle is defined for the entity.
      if (isset($entity_info['entity keys']['bundle'])) {
        $bundle_field = $entity_info['entity keys']['bundle'];

        // Extract list of bundles.
        $bundles = $search_api_bundle_settings['bundles'];

        // Extract operator.
        $bundle_operator = !empty($search_api_bundle_settings['default']) ? 'NOT IN' : 'IN';

        // Bundle is an entity.
        if (!isset($entity_properties[$bundle_field])  && !isset($entity_schema['fields'][$bundle_field]) && isset($entity_info['bundle keys']['bundle'])) {

          // Extract the entity property name corresponding to the bundle.
          $bundle_property_name = $entity_info['entity keys']['bundle'];
          // Extract the bundle key.
          $bundle_key = $entity_info['bundle keys']['bundle'];
          // Extract the bundle property.
          $bundle_property = trim(strchr($bundle_property_name, $bundle_key, TRUE), '_');

          // If the bundle property exists.
          if (isset($entity_properties[$bundle_property])) {
            // Get the bundle entity type and the bundle field.
            $bundle_entity_type = $entity_properties[$bundle_property]['type'];
            $bundle_field = $entity_properties[$bundle_property]['schema field'];

            // Load the bundle entities corresponding to the selected bundles.
            $query = new EntityFieldQuery();
            $result = $query->entityCondition('entity_type', $bundle_entity_type)
                            ->propertyCondition($bundle_key, $bundles, $bundle_operator)
                            ->execute();

            // Replace the bundles with the appropriate bundle field values.
            if (isset($result[$bundle_entity_type])) {
              $keys = array();

              $bundle_entities = entity_load($bundle_entity_type, array_keys($result[$bundle_entity_type]));
              foreach ($bundle_entities as $bundle_entity) {
                if (isset($bundle_entity->{$bundle_field})) {
                  $keys[] = $bundle_entity->{$bundle_field};
                }
              }

              if (count($keys) === count($bundles)) {
                $bundles = array_combine($keys, $bundles);
              }
              else {
                $bundles = NULL;
              }
            }
            else {
              $bundles = NULL;
            }
          }
          else {
            $bundles = NULL;
          }
        }

        // Prepare the bundle condition and entity bundle field.
        if (!empty($bundles)) {
          $bundle_condition_field = $bundle_operator . " ('" . implode("','", $bundles) . "')";
          $bundle_condition_entity = "{$entity_table}.{$bundle_field} {$bundle_operator} ('" . implode("','", array_keys($bundles)) . "')";
        }
      }
    }

    // Aggregated fields. Count type is not supported.
    $aggregated_fields = array();
    if (isset($index->options['data_alter_callbacks']['search_api_alter_add_aggregation']['settings']['fields'])) {
      $aggregated_fields = $index->options['data_alter_callbacks']['search_api_alter_add_aggregation']['settings']['fields'];
    }
    $aggregation_field_types = array(
      'fulltext' => array(
        'rt_field',
        'rt_attr_string',
        'rt_attr_uint',
        'rt_attr_timestamp',
        'rt_attr_float'),
      'sum' => array(
        'rt_attr_uint',
      ),
      'count' => array(),
      'max' => array(
        'rt_attr_uint',
        'rt_attr_timestamp',
        'rt_attr_float',
      ),
      'min' => array(
        'rt_attr_uint',
        'rt_attr_timestamp',
        'rt_attr_float',
      ),
      'first' => array(
        'rt_field',
        'rt_attr_string',
        'rt_attr_uint',
        'rt_attr_timestamp',
        'rt_attr_float',
      ),
    );

    $error = FALSE;

    // Parse Fields.
    foreach ($index_fields as $field_id => $field_data) {
      // Sphinx field info.
      $sphinx_field_name = $field_data['sphinx_field_name'];
      $sphinx_field_type = $field_data['sphinx_field_type'];
      $sphinx_field_type_sql = "sql" . drupal_substr($sphinx_field_type, 2);

      // Original field.
      $sql_field_id = isset($entity_properties[$field_id]['schema field']) ? $entity_properties[$field_id]['schema field'] : $field_id;

      // Entity field.
      if (isset($entity_schema['fields'][$sql_field_id])) {
        $query_main_fields[] = "{$entity_table}.{$sql_field_id} AS {$sphinx_field_name}";

        if ($sphinx_field_type !== 'rt_field') {
          $query_attr_simple[] = "{$sphinx_field_type_sql} = {$sphinx_field_name}";
        }
      }
      // Search API language field.
      elseif ($field_id === 'search_api_language') {
        if (isset($entity_properties['language'])) {
          $query_main_fields[] = "{$entity_table}.language AS {$sphinx_field_name}";
        }
        else {
          $query_main_fields[] = "'" . LANGUAGE_NONE . "' AS {$sphinx_field_name}";
        }

        $query_attr_simple[] = "{$sphinx_field_type_sql} = {$sphinx_field_name}";
      }
      // Aggregated field.
      elseif (isset($aggregated_fields[$field_id])) {
        $aggregated_field = $aggregated_fields[$field_id];
        $aggregation_field_type = $aggregation_field_types[$aggregated_field['type']];

        $fields = array();
        // We only handle aggregation of fields that are indexed.
        foreach ($aggregated_field['fields'] as $key => $sub_field_id) {
          if (isset($index_fields[$sub_field_id]) && in_array($index_fields[$sub_field_id]['sphinx_field_type'], $aggregation_field_type)) {
            $sql_sub_field_id = isset($entity_properties[$sub_field_id]['schema field']) ? $entity_properties[$sub_field_id]['schema field'] : $sub_field_id;

            if (isset($entity_schema['fields'][$sql_sub_field_id])) {
              $fields[] = "{$entity_table}.{$sql_sub_field_id}";
            }
            else {
              $sub_field_column = '';

              // Field with pattern: field_name:field_column.
              // We only handle 1 nested level (field attached to the entity).
              if (substr_count($sub_field_id, ':') === 1) {
                list($sub_field_id, $sub_field_column) = explode(':', $sub_field_id, 2);
              }

              // Get field info.
              $sub_field_info = field_info_field($sub_field_id);

              // We only handle fields using 'field_sql_storage'.
              if (!empty($sub_field_info) && isset($sub_field_info['storage']['type']) && $sub_field_info['storage']['type'] === 'field_sql_storage') {
                $sub_field_table = "field_data_{$sub_field_id}";

                if (empty($sub_field_column)) {
                  $sub_field_column = key($sub_field_info['columns']);
                }

                $fields[] = "{$sub_field_table}.{$sub_field_id}_{$sub_field_column}";
              }
            }
          }
        }

        // If we don't have the same number of fields after checking then
        // we can not proceed and we raise the error flag.
        if (count($fields) === count($aggregated_field['fields'])) {
          switch ($aggregated_field['type']) {
            case 'fulltext':
              $query_main_fields[] = sprintf("CONCAT_WS('\\n\\n', %s) AS %s", implode(', ', $fields), $sphinx_field_name);
              break;

            case 'sum':
              $query_main_fields[] = sprintf("SUM(%s) AS %s", implode(',', $fields), $sphinx_field_name);
              $query_attr_simple[] = "{$sphinx_field_type_sql} = {$sphinx_field_name}";
              break;

            case 'count':
              // Not supported.
              $error = TRUE;
              break;

            case 'max':
              $query_main_fields[] = sprintf("GREATEST(%s) AS %s", implode(',', $fields), $sphinx_field_name);
              $query_attr_simple[] = "{$sphinx_field_type_sql} = {$sphinx_field_name}";
              break;

            case 'min':
              $query_main_fields[] = sprintf("LEAST(%s) AS %s", implode(',', $fields), $sphinx_field_name);
              $query_attr_simple[] = "{$sphinx_field_type_sql} = {$sphinx_field_name}";
              break;

            case 'first':
              $query_main_fields[] = sprintf("CONCAT(COALESCE(%s)) AS %s", implode(', ', $fields), $sphinx_field_name);
              $query_attr_simple[] = "{$sphinx_field_type_sql} = {$sphinx_field_name}";
              break;
          }
        }
        else {
          $error = TRUE;
        }
      }
      // Field.
      else {
        $field_column = '';

        // Field with pattern: field_name:field_column.
        // We only handle 1 nested level so it's a field attached to the entity.
        if (substr_count($field_id, ':') === 1) {
          list($field_id, $field_column) = explode(':', $field_id, 2);
        }

        // Get field info.
        $field_info = field_info_field($field_id);

        // We only handle fields using 'field_sql_storage'.
        if (!empty($field_info) && isset($field_info['storage']['type']) && $field_info['storage']['type'] === 'field_sql_storage') {
          $field_table = "field_data_{$field_id}";

          // We assume the proper column is the first one
          // if no valid one is set.
          if (empty($field_column) || !isset($field_info['columns'][$field_column])) {
            $field_column = key($field_info['columns']);
          }

          $field_column = "{$field_id}_{$field_column}";

          if ($sphinx_field_type !== 'rt_attr_multi') {
            // Only join once.
            if (!isset($query_main_joins[$field_table])) {
              $join = array();
              $join[] = "LEFT JOIN {$field_table} AS {$field_table}";
              $join[] = "ON {$field_table}.entity_id = {$entity_table}.{$entity_id_field}";
              $join[] = "AND {$field_table}.entity_type = '{$entity_type}'";

              if (!empty($bundle_condition_field)) {
                $join[] = "AND {$field_table}.bundle {$bundle_condition_field}";
              }

              $query_main_joins[$field_table] = implode(" \\\n", $join);
            }

            $query_main_fields[] = "{$field_table}.{$field_column} AS {$sphinx_field_name}";

            if ($sphinx_field_type !== 'rt_field') {
              $query_attr_simple[] = "{$sphinx_field_type_sql} = {$sphinx_field_name}";
            }
          }
          else {
            $multi = array();
            $multi[] = "sql_attr_multi = uint {$sphinx_field_name} from ranged-query;";
            $multi[] = "SELECT {$entity_table}.{$entity_id_field}, IFNULL({$field_table}.{$field_column}, {$this->emptyMVA})";
            $multi[] = "FROM {$entity_table} AS {$entity_table}";
            $multi[] = "LEFT JOIN {$field_table} AS {$field_table}";
            $multi[] = "ON {$field_table}.entity_id = {$entity_table}.{$entity_id_field}";
            $multi[] = "AND {$field_table}.entity_type = '{$entity_type}'";

            if (!empty($bundle_condition_field)) {
              $multi[] = "AND {$field_table}.bundle {$bundle_condition_field}";
              $multi[] = "WHERE {$bundle_condition_entity}";
              $multi[] = "AND {$entity_table}.{$entity_id_field} >= {$start} AND {$entity_table}.{$entity_id_field} <= {$end};";
            }
            else {
              $multi[] = "WHERE {$entity_table}.{$entity_id_field} >= {$start} AND {$entity_table}.{$entity_id_field} <= {$end};";
            }

            $multi[] = "SELECT MIN({$entity_table}.{$entity_id_field}), MAX({$entity_table}.{$entity_id_field})";
            $multi[] = "FROM {$entity_table} AS {$entity_table}";
            if (!empty($bundle_condition_entity)) {
              $multi[] = "WHERE {$bundle_condition_entity}";
            }

            $query_attr_multi[] = implode(" \\\n", $multi);
          }
        }
        else {
          // ERROR: unknown field.
          $error = TRUE;
        }
      }

      if ($error === TRUE) {
        watchdog('search_api_sphinx', 'Fast indexing error. Unable to process field @field.', array('@field' => $field_id), WATCHDOG_ERROR);
        return;
      }
    }

    // Main query.
    $query_main = array();
    $query_main[] = "sql_query = SELECT";
    $query_main[] = implode(", \\\n", $query_main_fields);
    $query_main[] = "FROM {$entity_table} AS {$entity_table}";
    $query_main[] = implode(" \\\n", $query_main_joins);
    if (!empty($bundle_condition_entity)) {
      $query_main[] = "WHERE {$bundle_condition_entity}";
      $query_main[] = "AND {$entity_table}.{$entity_id_field} >= {$start} AND {$entity_table}.{$entity_id_field} <= {$end}";
    }
    else {
      $query_main[] = "WHERE {$entity_table}.{$entity_id_field} >= {$start} AND {$entity_table}.{$entity_id_field} <= {$end}";
    }
    $query_main = implode(" \\\n", $query_main);

    // Ranged query.
    $query_ranged = array();
    $query_ranged[] = "sql_query_range = SELECT MIN({$entity_table}.{$entity_id_field}), MAX({$entity_table}.{$entity_id_field})";
    $query_ranged[] = "FROM {$entity_table} AS {$entity_table}";
    if (!empty($bundle_condition_entity)) {
      $query_ranged[] = "WHERE {$bundle_condition_entity}";
    }
    $query_ranged = implode(" \\\n", $query_ranged);

    // Source configuration.
    $source_configuration = array();

    if ($credentials = drush_get_context('DRUSH_DB_CREDENTIALS')) {
      $sql_info = array();
      $sql_info[] = "type     = " . ($credentials['driver'] === 'mysqli' ? 'mysql' : $credentials['driver']);
      $sql_info[] = "sql_user = " . $credentials['user'];
      $sql_info[] = "sql_pass = " . $credentials['pass'];
      $sql_info[] = "sql_host = " . $credentials['host'];
      $sql_info[] = "sql_db   = " . $credentials['name'];
      $sql_info[] = "sql_port = " . (isset($credentials['port']) ? $credentials['port'] : '3306');

      $source_configuration[] = implode("\n", $sql_info);
    }

    $source_configuration[] = "sql_query_pre = SET NAMES utf8";
    $source_configuration[] = $query_ranged;
    $source_configuration[] = "sql_range_step = 1000";
    $source_configuration[] = "sql_ranged_throttle = 0";
    $source_configuration[] = $query_main;
    $source_configuration[] = implode("\n", $query_attr_simple);
    $source_configuration[] = implode("\n\n", $query_attr_multi);

    $source_configuration = "source {$index_name}_sql\n{\n" . implode("\n\n", $source_configuration) . "\n}\n";

    // Index configuration.
    $index_configuration = array();

    $index_options = array(
      'source' => "{$index_name}_sql",
      'path' => $this->getSphinxServerDirectory() . '/' . $index_name . '_temp',
      'dict' => 'keywords',
    );

    $index_options += $this->getIndexOptions($index, 'index');

    foreach ($index_options as $option => $value) {
      // Only add options that are not specific to RT indexes.
      if (drupal_substr($option, 0, 3) !== 'rt_') {
        $value = str_replace(array("\r", "\n"), array('', "\\\n\t"), $value);
        $index_configuration[] = "{$option} = {$value}";
      }
    }

    $index_configuration = "index {$index_name}_temp\n{\n" . implode("\n", $index_configuration) . "\n}\n";

    // Indexer configuration.
    $indexer_configuration = array();
    $indexer_configuration[] = "mem_limit         = " . $this->getIndexOption($index, 'index', 'rt_mem_limit', '256M');
    $indexer_configuration[] = "max_iops          = 0";
    $indexer_configuration[] = "max_iosize        = 0";
    $indexer_configuration[] = "write_buffer      = 4M";
    $indexer_configuration = "indexer\n{\n" . implode("\n", $indexer_configuration) . "\n}\n";

    // Main configuration.
    $configuration = "{$source_configuration}\n{$index_configuration}\n{$indexer_configuration}";

    // First we stop the daemon.
    // All indexes for this server are flushed to disk.
    $this->startStopSphinxDaemon('stop');

    // Remove connection to sphinx daemon.
    $this->removeSphinxConnection();

    // Then we delete the index files for the given index(es).
    $this->deleteSphinxIndexFiles($index);

    // Update the server directory if necessary.
    $updated = $this->updateSphinxServerDirectory();

    // We recreate the server configuration file with the extra configuration.
    $rebuilt = $this->updateSphinxServerConfigurationFile($configuration);

    if ($rebuilt) {
      $command = "indexing";

      $indexer = $this->getOption('indexer', '/usr/bin/indexer');

      $configuration_file = $this->getSphinxServerConfigurationFile();

      $indexer_command = "{$indexer} --config \"{$configuration_file}\" --noprogress {$index_name}_temp";

      $output = array();

      // Remove indexing cron tasks.
      _search_api_empty_cron_queue($index);

      // Start the indexing.
      exec($indexer_command, $output, $return_val);

      // Reconnect to the Drupal database as connection may have ended.
      Database::closeConnection();
      Database::getConnection();

      watchdog('search_api_sphinx_indexer', "@action\nCommand: @command\nReturn value: @return_val\n\nOutput: @output", array(
        '@action' => drupal_strtoupper($command),
        '@command' => $indexer_command,
        '@return_val' => $return_val,
        '@output' => implode("\n", (array) $output),
      ));

      if (empty($return_val)) {
        // And finally we restart the daemon if configuration was written.
        if (!$this->startStopSphinxDaemon('start')) {
          throw new SearchApiException(t('Error while starting the sphinx daemon. Please read the log for more information.'));
        }
        else {
          $datasource = $index->datasource();

          // Attach the disk index to the RT index.
          $this->querySphinx("ATTACH INDEX {$index_name}_temp TO RTINDEX {$index_name}")->rowCount();

          // Get the number of items indexed by Sphinx.
          $this->querySphinx("SELECT * FROM {$index_name} LIMIT 0")->fetchField();
          $result_count = $this->getSphinxQueryMeta('total_found');

          // Check if the attach action succeeded.
          if (!empty($result_count)) {
            // Update the datasource.
            while (TRUE) {
              $item_ids = $datasource->getChangedItems($index, 1000);
              if (!empty($item_ids)) {
                $index->datasource()->trackItemIndexed($item_ids, $index);
              }
              else {
                break;
              }
            }
          }
          else {
            $return_val = 1;
          }
        }
      }

      // Update the sphinx server to use the original configuration.
      $this->updateSphinxServer();

      return empty($return_val);
    }

    return FALSE;
  }
}
